/* Author: tmouse a!t bigfoot d!o!t com */
/* Notes: Hashing is useful in cryptography (to hash passwords rather than storing
 * in plain text, for example) and for data (such as files) equality testing.
 * Windows security provides an API for data hashing.
 * Beware that the hash functions work on binary data. Therefore, the hash of
 * the unicide version of a string will be different to the hash of the ansi version. 
 */


#include <windows.h>
#include <wincrypt.h>
#pragma comment(lib, "advapi32.lib")

/*
 * Creates a hexadecimal hash string from input data.
 * ALG_ID can be one of CALG_MD2, CALG_MD4, CALG_MD5 or CALG_SHA1.
 * SHA1 algorithm is strongest, while MD2 and MD4 should be avoided.
 * MD algorithms require a 33 character output buffer. SHA1 requires 41 characters.
 * Returns TRUE on success, FALSE on failure.
 */
BOOL HashData(LPTSTR szOut, DWORD cchOut, LPCBYTE lpIn, DWORD cbIn, ALG_ID hash_algorithm)
{
	HCRYPTPROV hProv      = 0;
	HCRYPTHASH hHash      = 0;
	BYTE       raw[64];
	DWORD      raw_len    = sizeof(raw);
	BOOL       bResult    = FALSE;
	UINT       i;

	if(CryptAcquireContext(&hProv, NULL, NULL, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT | CRYPT_SILENT) &&
	   CryptCreateHash(hProv, hash_algorithm, 0, 0, &hHash) &&
	   CryptHashData(hHash, lpIn, cbIn, 0) &&
	   CryptGetHashParam(hHash, HP_HASHVAL, raw, &raw_len, 0) &&
	   (raw_len * 2) + 1 <= cchOut)
	{
		for (i = 0;i < raw_len;i++)
		{
			wsprintf(&szOut[i * 2], TEXT("%02.2x"), raw[i]);
		}

		bResult = TRUE;
	}

	if (hHash) CryptDestroyHash(hHash);
	if (hProv) CryptReleaseContext(hProv, 0);

	return bResult;
}


#if 1 /* Test code. */
#include <stdio.h>
#include <tchar.h>

int main(void)
{
	TCHAR szHash[64];
	char * test = "Hello, this is a test";

	printf("Input: %s\n", test);

	if (HashData(szHash, 33, test, strlen(test), CALG_MD5))
		_tprintf(_T("MD5: %s\n"), szHash);

	if (HashData(szHash, 33, test, strlen(test), CALG_MD2))
		_tprintf(_T("MD2: %s\n"), szHash);

	if (HashData(szHash, 33, test, strlen(test), CALG_MD4))
		_tprintf(_T("MD4: %s\n"), szHash);

	if (HashData(szHash, 41, test, strlen(test), CALG_SHA1))
		_tprintf(_T("SHA1: %s\n"), szHash);

	getchar();
	return 0;
}
#endif


/* =================================================================== */
/* If you have a large amount of data, you may want to hash it in chunks. */

/* 
 * Create the hash.
 */
HCRYPTHASH HashInit(HCRYPTPROV hCryptProvider, ALG_ID hash_algorithm)
{
	HCRYPTHASH hHash = 0;
	CryptCreateHash(hCryptProvider, hash_algorithm, 0, 0, &hHash);
	return hHash;
}

/*
 * Add data to the hash.
 */
BOOL HashUpdate(HCRYPTHASH hHash, LPCBYTE lpData, DWORD cbData)
{
	return CryptHashData(hHash, lpData, cbData, 0);
}

/* 
 * Return the hash hexadecimal value and destroy the hash.
 */
BOOL HashFinal(HCRYPTHASH hHash, LPTSTR szOut, DWORD cchOut)
{
	BYTE       raw[64];
	DWORD      raw_len    = sizeof(raw);
	BOOL       bResult    = FALSE;
	UINT       i;

	if (CryptGetHashParam(hHash, HP_HASHVAL, raw, &raw_len, 0) &&
	    (raw_len * 2) + 1 <= cchOut)
	{
		for (i = 0;i < raw_len;i++)
		{
			wsprintf(&szOut[i * 2], TEXT("%02.2x"), raw[i]);
		}

		bResult = TRUE;
	}

	if (hHash) CryptDestroyHash(hHash);

	return bResult;
}


#if 0 /* Test code. */
#include <stdio.h>
#include <tchar.h>

int main(void)
{
	HCRYPTPROV hProv      = 0;
	HCRYPTHASH hHash      = 0;
	TCHAR szHash[64];

	CryptAcquireContext(&hProv, NULL, NULL, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT | CRYPT_SILENT);

	hHash = HashInit(hProv, CALG_MD5);
	HashUpdate(hHash, "abcd", strlen("abcd"));
	HashUpdate(hHash, "efgh", strlen("efgh"));
	HashFinal(hHash, szHash, 33);
	_tprintf(_T("MD5: %s\n"), szHash);

	hHash = HashInit(hProv, CALG_SHA1);
	HashUpdate(hHash, "abcd", strlen("abcd"));
	HashUpdate(hHash, "efgh", strlen("efgh"));
	HashFinal(hHash, szHash, 41);
	_tprintf(_T("SHA1: %s\n"), szHash);

	CryptReleaseContext(hProv, 0);
	getchar();
	return 0;
}
#endif

